#!/usr/bin/python
#emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*- 
#ex: set sts=4 ts=4 sw=4 noet:
#------------------------- =+- Python script -+= -------------------------
"""
 @file      common.py
 @date      Tue Jun  7 09:22:57 2011
 @brief


  Yaroslav Halchenko                                            Dartmouth
  web:     http://www.onerussian.com                              College
  e-mail:  yoh@onerussian.com                              ICQ#: 60653192

 DESCRIPTION (NOTES):

 COPYRIGHT: Yaroslav Halchenko 2011

 LICENSE: MIT

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""
#-----------------\____________________________________/------------------

__author__ = 'Yaroslav Halchenko'
__copyright__ = 'Copyright (c) 2011 Yaroslav Halchenko'
__license__ = 'MIT'

# Each tuple here is a (regular expression for "Others", either it was newly added while survey was on already)
entries_to_refresh = dict(
    sw_other_name=dict(
        sw_electro=dict(
            brainsuite=('brainsuite', True),                # EARLYIN
            cedspike=('ced *spike2*', True),                # NEW: http://www.ced.co.uk/pru.shtml?spk4wglu.htm
            datariver=('exp control: datariver', True),     # NEW: http://sccn.ucsd.edu/wiki/DataSuite
            eeglab=('(eeglab|http://sccn.ucsd.edu/eeglab/)', False),
            emse=('emse', True),                            # REFRESH  EARLYIN
            erplab=('erplab', True),                        # NEW:     ERPLAB
            klusters=('klusters.*', True),                  # REFRESH
            netstation=('egi net station', True),           # NEW:     EGI Net Station
            neuroscan=('(curry|neuroscan(| curry))', True), # REFRESH
            neuroscope=('.*neuroscope', True),              # REFRESH
            nutmeg=('.*nutmeg', True),                      # NEW
            ndmanager=('ndmanager', True),                  # EARLYIN
            fmatoolbox=('FMAToolbox', True),                # EARLYIN
            ),
        sw_img=dict(
            mricron=('mricrogl', False),
            afni=('afni for bci', False),
            dtistudio=('dti-*studio', True),    # NEW: or MRIStudio?
            brainsight=('brainsight', True),    # NEW: BrainSight
            nordicice=('nordic ice', True),     # NEW: NordicICE  -- just 1
            trackvis=('trackvis', False),
            xmedcon=('xmedcon', True),          # NEW
            ),
        sw_general=dict(
            lua=('lua', True),                  # NEW
            stata=('stata', True),              # NEW
            statistica=('statistica', True),    # NEW
            java=('java', True),                # REFRESH
            fortran=('fortran', True),          # EARLYIN
            ),
        sw_neusys=dict(
            ecanse=('ecanse', True),            # EARLYIN
            emergent=('emergent', True),        # EARLYIN
            nengo=('nengo', True),              # EARLYIN
            neuroml=('neuroml', True),          # NEW: NeuroML -- more of a framework/standard than software
            neurosolutions=('neurosolutions', True), # EARLYIN
            peltarion=('peltarion', True),           # EARLYIN
            snnap=('snnap', True),              # EARLYIN
            snns=('snns', True),                # EARLYIN
            xnbc=('xnbc', True),                # EARLYIN
            xpp=('xpp(|y|aut)', True),          # REFRESH: XPP/XPPAUT and Python interface
            ),
        sw_psychphys=dict(
            asf=('asf', True),                  # NEW: ASF  http://code.google.com/p/asf/
            cogent=('cogent(|2000)', True),     # REFRESH
            crsvsg=('crs toolbox.*', True),     # NEW: CRS VSG Toolbox  http://www.crsltd.com/catalog/vsgtoolbox/
            mindware=('mind-ware', True),       # NEW: MindWare
            nordicaktiva=('nordic aktiva', True), # NEW:    NordicActiva  -- just 1 http://www.nordicneurolab.com/Products_and_Solutions/Clinical_Software_Solutions/nordicActiva.aspx  http://www.nordicneurolab.com/Products_and_Solutions/Clinical_Software_Solutions/nordicAktiva.aspx
            superlab=('superlab', True),        # REFRESH
            psignifit=('psignifit(|3)', True),  # NEW
            ),
        ignore=dict(ignore=(
                    '(zsh vim mutt git'
                    # just ignore
                    '|my overall time.*|separate work.*|60% windows'
                    '|.*my own .*software'
                    # Different generic visualization solutions
                    '|gnupot|.*gnu plot.*xmgrace|mayavi|matplotlib'
                    '|trackvis'
                    '|opengl|itk|vtk'
                    '|paraview'
                    # Really cool one for graphs
                    '|gephi'
                    # Generic DBs
                    '|mysql|postgresql'
                    # DB with imaging data (Italy?) but just once
                    '|loris multi-site database system'
                    # More languages/platforms?
                    '|.net|haskel|gsl|cuda'
                    # Python lovers
                    '|theano|pygame|numpy|mdp|joblib|scipy|pytables|sympy'
                    # ML toolboxes
                    '|scikits-learn|probid .*'
                    # Reference managers
                    '|mendeley|jabref'
                    # Python IDE?? quite nice btw
                    '|spyder'
                    # Move into survey?
                    '|.*magnetic source locator.*' # Some kind of MEG inverse solver -- publications but no public project
                    ')', False)
                    ),
        ),
    )



# −*− coding: UTF−8 −*−
from path import path
import os
import pickle
"""
A class providing dictionary access to a folder.
cribbed from http://bitbucket.org/howthebodyworks/fsdict
"""

def get_tmp_dir():
    import tempfile
    return tempfile.mkdtemp()
    
class FSDict(dict):
    """
    provide dictionary access to a temp dir. I don't know why i didn't just use
    shelve. I think I forgot it existed.
    
    N.B. the keys ordering here is FS-dependent and thus unlike to be the same as
    with a real dict. beware.
    """
    
    unclean_dirs = []
    
    def __init__(self, initval=[], work_dir=None, *args, **kwargs):
        if work_dir is None:
            work_dir = get_tmp_dir()
        self.work_dir = path(work_dir)
        if not self.work_dir.exists():
            self.work_dir.mkdir()
        for key, val in getattr(initval, 'iteritems', initval.__iter__)():
            self[key] = val
        self.unclean_dirs.append(self.work_dir)
        super(FSDict, self).__init__(*args, **kwargs)
    
    def __setitem__(self, key, val, *args, **kwargs):
        pickle.dump(val, open(self.work_dir/key, 'w'))
    
    def __getitem__(self, key, *args, **kwargs):
        return pickle.load(open(self.work_dir/key, 'r'))
    
    def __repr__(self):
        """
        a hardline list of everything in the dict. may be long.
        """
        return repr(dict([(k, v) for k, v in self.iteritems()]))
        
    def __str__(self):
        """
        str is truncated somewhat.
        """
        if len(self.keys()):
            return '{' + repr(self.keys()[0]) + ':' + repr(self[self.keys()[0]]) + ', ...'
        else:
            return super(FSDict, self).__str__()
    
    def keys(self, *args, **kwargs):
        return [key for key in self.iterkeys()]
    
    def iterkeys(self, *args, **kwargs):
        for f in self.work_dir.files():
            yield str(self.work_dir.relpathto(f))
        
    def iteritems(self):
        for key in self.iterkeys():
            yield key, self[key]
            
    def itervalues(self):
        for key in self.iterkeys():
            yield self[key]
            
    def __delitem__(self, key, *args, **kwargs):
        (self.work_dir/key).unlink()
    
    def values(self, *args, **kwargs):
        return [self[key] for key in self.keys()]
        
    def cleanup(self):
        self.work_dir.rmtree()
    
    @classmethod
    def cleanup_all(cls):
        for fsd in cls.unclean_dirs:
            try:
                fsd.rmtree()
            except OSError:
                pass
    
    def move(self, new_dir):
        
        try:
            self.work_dir.move(new_dir)
        except Exception, e:
            raise
        else:
            self.work_dir = new_dir
    
    def __eq__(self, other):
        """
        when compared to a dict, equate equal if all keys and vals are equal
        note, this is potentially expensive.
        """
        #duck type our way to sanity:
        if not hasattr(other, 'keys'): return False
        #OK, it's a dict-ish thing
        try:
            return all([self[key]==other[key] for key in other]) and \
              len(self.keys())==len(other.keys())
        except KeyError:
            return False


# -*- coding: utf-8 -*-
from fsdict import FSDict
import feedgenerator
from urllib import quote_plus
import os.path
import re

#global
feed_entries = None

#constant unlikely to occur in a docname and legal as a filename
MAGIC_SEPARATOR = '---###---'

def parse_date(datestring):
    try:
        parser = parse_date.parser
    except AttributeError:
        import dateutil.parser
        parser = dateutil.parser.parser()
        parse_date.parser = parser
    return parser.parse(datestring)
    
def setup(app):
    """
    see: http://sphinx.pocoo.org/ext/appapi.html
    this is the primary extension point for Sphinx
    """
    from sphinx.application import Sphinx
    if not isinstance(app, Sphinx): return
    app.add_config_value('feed_base_url', '', 'html')
    app.add_config_value('feed_description', '', 'html')
    app.add_config_value('feed_filename', 'rss.xml', 'html')
    app.add_config_value('feed_title', '', 'html')
    app.add_config_value('feed_subtitle', '', 'html')
    app.add_config_value('feed_author_name', '', 'html')
    app.add_config_value('feed_author_email', '', 'html')
    app.add_config_value('feed_categories', [], 'html')
    app.add_config_value('feed_variants',
                         {'all': {'filename': 'rss.xml', 'categories': None}},
                         'html')
    app.connect('html-page-context', create_feed_item)
    app.connect('build-finished', emit_feed)
    app.connect('builder-inited', create_feed_container)
    app.connect('env-purge-doc', remove_dead_feed_item)
    
def create_feed_container(app):
    """
    create lazy filesystem stash for keeping RSS entry fragments, since we
    don't want to store the entire site in the environment (in fact, even if
    we did, it wasn't persisting for some reason.)
    """
    global feed_entries
    rss_fragment_path = os.path.realpath(os.path.join(app.outdir, '..', 'rss_entry_fragments'))
    feed_entries = FSDict(work_dir=rss_fragment_path)
    app.builder.env.feed_url = app.config.feed_base_url + '/' + \
        app.config.feed_filename

def inject_feed_url(app, pagename, templatename, ctx, doctree):
    #We like to provide our templates with a way to link to the rss output file
    ctx['rss_link'] = app.builder.env.feed_url #app.config.feed_base_url + '/' + app.config.feed_filename
    

def create_feed_item(app, pagename, templatename, ctx, doctree):
    """
    Here we have access to nice HTML fragments to use in, say, an RSS feed.
    We serialize them to disk so that we get them preserved across builds.
    
    We also inject useful metadata into the context here.
    """
    global feed_entries
    from absolutify_urls import absolutify
    metadata = app.builder.env.metadata.get(pagename, {})
    
    if 'date' not in metadata:
        return #don't index dateless articles
    try:
        pub_date = parse_date(metadata['date'])
        app.builder.env.metadata.get(pagename, {})
    except ValueError, exc:
        #probably a nonsensical date
        app.builder.warn('date parse error: ' + str(exc) + ' in ' + pagename)
        return
    
    # RSS item attributes, w/defaults:
    #     title, link, description, author_email=None,
    #     author_name=None, author_link=None, pubdate=None, comments=None,
    #     unique_id=None, enclosure=None, categories=(), item_copyright=None,
    #     ttl=None,
    link = app.config.feed_base_url + '/' + ctx['current_page_name'] + ctx['file_suffix']
    # bring main body of the feed item into shape
    body = ctx.get('body')
    # remove all header links (they make ugly characters in feed readers)
    body = re.sub('\<a class\="headerlink".*\>.</a\>', '', body)

    item = {
      'title': ctx.get('title'),
      'link': link,
      'unique_id': link,
      'description': absolutify(body, link),
      'pubdate': pub_date,
      'categories': ()
    }
    if 'tags' in metadata:
        item['categories'] = metadata['tags'].split(",")
    if 'author' in metadata:
        item['author_name'] = metadata['author']
    else:
        item['author_name'] = app.config.feed_author_name
    if 'author_email' in metadata:
        item['author_email'] = metadata['author_email']
    else:
        item['author_email'] = app.config.feed_author_email
    feed_entries[nice_name(pagename, pub_date)] = item    
    
    #Now, useful variables to keep in context
    ctx['rss_link'] = app.builder.env.feed_url 
    ctx['pub_date'] = pub_date

def remove_dead_feed_item(app, env, docname):
    """
    TODO:
    purge unwanted crap
    """
    global feed_entries
    munged_name = ''.join([MAGIC_SEPARATOR,quote_plus(docname)])
    for name in feed_entries:
        if name.endswith(munged_name):
            del(feed_entries[name])

def emit_feed(app, exc):
    global feed_entries
    import os.path
    
    title = app.config.feed_title
    if not title:
        title = app.config.project

    feed_dict = {
      'title': title,
      'subtitle': app.config.feed_subtitle,
      'link': app.config.feed_base_url,
      'feed_url': app.config.feed_base_url,
      'description': app.config.feed_description,
      'categories': app.config.feed_categories,
      'author_name': app.config.feed_author_name,
      'author_email': app.config.feed_author_email
    }
    if app.config.language:
        feed_dict['language'] = app.config.language
    if app.config.copyright:
        feed_dict['feed_copyright'] = app.config.copyright
    # sort items
    ordered_keys = feed_entries.keys()
    ordered_keys.sort(reverse=True)
    # loop over all feed variants
    for feedvar in app.config.feed_variants:
        feedvar_settings = app.config.feed_variants[feedvar]
        feed = feedgenerator.Rss201rev2Feed(**feed_dict)
        app.builder.env.feed_feed = feed
        for key in ordered_keys:
            item = feed_entries[key]
            # only take the ones that should be in this feed
            if feedvar_settings['tag'] is None \
                    or feedvar_settings['tag'] in item['categories']:
                feed.add_item(**feed_entries[key])
        outfilename = os.path.join(app.builder.outdir,
          feedvar_settings['filename'])
        # make sure the directory exists
        feed_dir = os.path.dirname(outfilename)
        if feed_dir and not os.path.exists(feed_dir):
            os.makedirs(os.path.dirname(outfilename))
        fp = open(outfilename, 'w')
        feed.write(fp, 'utf-8')
        fp.close()

def nice_name(docname, date):
    """
    we need convenient filenames which incorporate dates for ease of sorting and
    guid for uniqueness, plus will work in the FS without inconvenient
    characters. NB, at the moment, hour of publication is ignored.
    """
    return quote_plus(MAGIC_SEPARATOR.join([date.isoformat(), docname]))



# By Gareth Rees
# http://gareth-rees.livejournal.com/27148.html

import html5lib
import html5lib.serializer
import html5lib.treewalkers
import urlparse

# List of (ELEMENT, ATTRIBUTE) for HTML5 attributes which contain URLs.
# Based on the list at http://www.feedparser.org/docs/resolving-relative-links.html
url_attributes = [
    ('a', 'href'),
    ('applet', 'codebase'),
    ('area', 'href'),
    ('blockquote', 'cite'),
    ('body', 'background'),
    ('del', 'cite'),
    ('form', 'action'),
    ('frame', 'longdesc'),
    ('frame', 'src'),
    ('iframe', 'longdesc'),
    ('iframe', 'src'),
    ('head', 'profile'),
    ('img', 'longdesc'),
    ('img', 'src'),
    ('img', 'usemap'),
    ('input', 'src'),
    ('input', 'usemap'),
    ('ins', 'cite'),
    ('link', 'href'),
    ('object', 'classid'),
    ('object', 'codebase'),
    ('object', 'data'),
    ('object', 'usemap'),
    ('q', 'cite'),
    ('script', 'src')]

def absolutify(src, base_url):
    """absolutify(SRC, BASE_URL): Resolve relative URLs in SRC.
SRC is a string containing HTML. All URLs in SRC are resolved relative
to BASE_URL. Return the body of the result as HTML."""

    # Parse SRC as HTML.
    tree_builder = html5lib.treebuilders.getTreeBuilder('dom')
    parser = html5lib.html5parser.HTMLParser(tree = tree_builder)
    dom = parser.parse(src)

    # Handle <BASE> if any.
    head = dom.getElementsByTagName('head')[0]
    for b in head.getElementsByTagName('base'):
        u = b.getAttribute('href')
        if u:
            base_url = urlparse.urljoin(base_url, u)
            # HTML5 4.2.3 "if there are multiple base elements with href
            # attributes, all but the first are ignored."
            break

    # Change all relative URLs to absolute URLs by resolving them
    # relative to BASE_URL. Note that we need to do this even for URLs
    # that consist only of a fragment identifier, because Google Reader
    # changes href=#foo to href=http://site/#foo
    for tag, attr in url_attributes:
        for e in dom.getElementsByTagName(tag):
            u = e.getAttribute(attr)
            if u:
                e.setAttribute(attr, urlparse.urljoin(base_url, u))

    # Return the HTML5 serialization of the <BODY> of the result (we don't
    # want the <HEAD>: this breaks feed readers).
    body = dom.getElementsByTagName('body')[0]
    tree_walker = html5lib.treewalkers.getTreeWalker('dom')
    html_serializer = html5lib.serializer.htmlserializer.HTMLSerializer()
    return u''.join(html_serializer.serialize(tree_walker(body)))
    

# Alternative option, from http://stackoverflow.com/questions/589833/how-to-find-a-relative-url-and-translate-it-to-an-absolute-url-in-python/589939#589939
# 
# import re, urlparse
# 
# find_re = re.compile(r'\bhref\s*=\s*("[^"]*"|\'[^\']*\'|[^"\'<>=\s]+)')
# 
# def fix_urls(document, base_url):
#     ret = []
#     last_end = 0
#     for match in find_re.finditer(document):
#         url = match.group(1)
#         if url[0] in "\"'":
#             url = url.strip(url[0])
#         parsed = urlparse.urlparse(url)
#         if parsed.scheme == parsed.netloc == '': #relative to domain
#             url = urlparse.urljoin(base_url, url)
#             ret.append(document[last_end:match.start(1)])
#             ret.append('"%s"' % (url,))
#             last_end = match.end(1)
#     ret.append(document[last_end:])
#     return ''.join(ret)
